/*
  Remember row,column where string was acquired.
  postcombobox x y 
	icon?, text, row, column position, fg, bg button?
	based upon style.
  grab set
  SetIcon
  SetText
  SetBg
  SetFg
  SetFont
  SetButton  
 */

/*
 * bltTreeViewEdit.c --
 *
 *	This module implements an hierarchy widget for the BLT toolkit.
 *
 * Copyright 1998-1999 Lucent Technologies, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that the copyright notice and warranty
 * disclaimer appear in supporting documentation, and that the names
 * of Lucent Technologies or any of their entities not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.
 *
 * Lucent Technologies disclaims all warranties with regard to this
 * software, including all implied warranties of merchantability and
 * fitness.  In no event shall Lucent Technologies be liable for any
 * special, indirect or consequential damages or any damages
 * whatsoever resulting from loss of use, data or profits, whether in
 * an action of contract, negligence or other tortuous action, arising
 * out of or in connection with the use or performance of this
 * software.
 *
 *	The "treeview" widget was created by George A. Howlett.
 */

#include "bltInt.h"

#ifndef NO_TREEVIEW

#include "bltTreeView.h"
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#define COMBOBOX_FOCUS	(1<<0)
#define COMBOBOX_REDRAW	(1<<1)

static Tcl_IdleProc DisplayCombobox;
static Tcl_FreeProc DestroyCombobox;
static Tcl_TimerProc BlinkCursorProc;
static Tcl_ObjCmdProc ComboboxCmd;

/*
 * Combobox --
 *
 *	This structure is shared by entries when their labels are
 *	edited via the keyboard.  It maintains the location of the
 *	insertion cursor and the text selection for the editted entry.
 *	The structure is shared since we need only one.  The "focus"
 *	entry should be the only entry receiving KeyPress/KeyRelease
 *	events at any time.  Information from the previously editted
 *	entry is overwritten.
 *
 *	Note that all the indices internally are in terms of bytes,
 *	not characters.  This is because UTF-8 strings may encode a
 *	single character into a multi-byte sequence.  To find the
 *	respective character position
 *
 *		n = Tcl_NumUtfChars(string, index);
 *
 *	where n is the resulting character number.
 */
typedef struct {

    /*
     * This is a SNAFU in the Tk API.  It assumes that only an official
     * Tk "toplevel" widget will ever become a toplevel window (i.e. a
     * window whose parent is the root window).  Because under Win32,
     * Tk tries to use the widget record associated with the TopLevel
     * as a Tk frame widget, to read its menu name.  What this means
     * is that any widget that's going to be a toplevel, must also look
     * like a frame. Therefore we've copied the frame widget structure
     * fields into the token.
     */

    Tk_Window tkwin;		/* Window that embodies the frame.  NULL
				 * means that the window has been destroyed
				 * but the data structures haven't yet been
				 * cleaned up. */
    Display *display;		/* Display containing widget.  Used, among
				 * other things, so that resources can be
				 * freed even after tkwin has gone away. */
    Tcl_Interp *interp;		/* Interpreter associated with widget.  Used
				 * to delete widget command. */
    Tcl_Command widgetCmd;	/* Token for frame's widget command. */
    char *className;		/* Class name for widget (from configuration
				 * option).  Malloc-ed. */
    int mask;			/* Either FRAME or TOPLEVEL;  used to select
				 * which configuration options are valid for
				 * widget. */
    char *screenName;		/* Screen on which widget is created.  Non-null
				 * only for top-levels.  Malloc-ed, may be
				 * NULL. */
    char *visualName;		/* Textual description of visual for window,
				 * from -visual option.  Malloc-ed, may be
				 * NULL. */
    char *colormapName;		/* Textual description of colormap for window,
				 * from -colormap option.  Malloc-ed, may be
				 * NULL. */
    char *menuName;		/* Textual description of menu to use for
				 * menubar. Malloc-ed, may be NULL. */
    Colormap colormap;		/* If not None, identifies a colormap
				 * allocated for this window, which must be
				 * freed when the window is deleted. */
    Tk_3DBorder border;		/* Structure used to draw 3-D border and
				 * background.  NULL means no background
				 * or border. */
    int borderWidth;		/* Width of 3-D border (if any). */
    int relief;			/* 3-d effect: TK_RELIEF_RAISED etc. */
    int highlightWidth;		/* Width in pixels of highlight to draw
				 * around widget when it has the focus.
				 * 0 means don't draw a highlight. */
    XColor *highlightBgColorPtr;
				/* Color for drawing traversal highlight
				 * area when highlight is off. */
    XColor *highlightColorPtr;	/* Color for drawing traversal highlight. */
    int width;			/* Width to request for window.  <= 0 means
				 * don't request any size. */
    int height;			/* Height to request for window.  <= 0 means
				 * don't request any size. */
    Tk_Cursor cursor;		/* Current cursor for window, or None. */
    char *takeFocus;		/* Value of -takefocus option;  not used in
				 * the C code, but used by keyboard traversal
				 * scripts.  Malloc'ed, but may be NULL. */
    int isContainer;		/* 1 means this window is a container, 0 means
				 * that it isn't. */
    char *useThis;		/* If the window is embedded, this points to
				 * the name of the window in which it is
				 * embedded (malloc'ed).  For non-embedded
				 * windows this is NULL. */
    int flags;			/* Various flags;  see below for
				 * definitions. */

    /* Combobox-specific fields */
    TreeView *tvPtr;
    int x, y;			/* Position of window. */

    int active;			/* Indicates that the frame is active. */
    int exportSelection;

    int insertPos;		/* Position of the cursor within the
				 * array of bytes of the entry's label. */

    int cursorX, cursorY;	/* Position of the insertion cursor in the
				 * combobox window. */
    short int cursorWidth;	/* Size of the insertion cursor symbol. */
    short int cursorHeight;

    int selAnchor;		/* Fixed end of selection. Used to extend
				 * the selection while maintaining the
				 * other end of the selection. */
    int selFirst;		/* Position of the first character in the
				 * selection. */
    int selLast;		/* Position of the last character in the
				 * selection. */

    int cursorOn;		/* Indicates if the cursor is displayed. */
    int onTime, offTime;	/* Time in milliseconds to wait before
				 * changing the cursor from off-to-on
				 * and on-to-off. Setting offTime to 0 makes
				 * the cursor steady. */
    Tcl_TimerToken timerToken;	/* Handle for a timer event called periodically
				 * to blink the cursor. */
    /* Data-specific fields. */
    TreeViewEntry *entryPtr;	/* Selected entry */
    TreeViewColumn *columnPtr;	/* Column of entry to be edited */
    TreeViewStyle *stylePtr;
    TreeViewIcon icon;
    int gap;
    char *string;
    TextLayout *textPtr;
    Tk_Font font;
    GC gc;

    Tk_3DBorder selBorder;
    int selRelief;
    int selBorderWidth;
    XColor *selFgColor;		/* Text color of a selected entry. */
} Combobox;

#define DEF_COMBOBOX_BACKGROUND		RGB_WHITE
#define DEF_COMBOBOX_BORDER_WIDTH		STD_BORDERWIDTH
#define DEF_COMBOBOX_CURSOR		(char *)NULL
#define DEF_COMBOBOX_EXPORT_SELECTION	"no"
#define DEF_COMBOBOX_NORMAL_BG_COLOR 	STD_COLOR_NORMAL_BG
#define DEF_COMBOBOX_NORMAL_FG_MONO	STD_MONO_ACTIVE_FG
#define DEF_COMBOBOX_RELIEF		"sunken"
#define DEF_COMBOBOX_SELECT_BG_COLOR 	RGB_LIGHTBLUE0
#define DEF_COMBOBOX_SELECT_BG_MONO  	STD_MONO_SELECT_BG
#define DEF_COMBOBOX_SELECT_BORDER_WIDTH	"1"
#define DEF_COMBOBOX_SELECT_FG_COLOR 	STD_COLOR_SELECT_FG
#define DEF_COMBOBOX_SELECT_FG_MONO  	STD_MONO_SELECT_FG
#define DEF_COMBOBOX_SELECT_RELIEF	"flat"

/* Combobox Procedures */
static Blt_ConfigSpec comboboxConfigSpecs[] =
{
    {BLT_CONFIG_BORDER, "-background", "background", "Background",
	DEF_COMBOBOX_BACKGROUND, Blt_Offset(Combobox, border), 0},
    {BLT_CONFIG_SYNONYM, "-bd", "borderWidth", (char *)NULL, (char *)NULL, 0,0},
    {BLT_CONFIG_SYNONYM, "-bg", "background", (char *)NULL, (char *)NULL, 0,0},
    {BLT_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
	DEF_COMBOBOX_CURSOR, Blt_Offset(Combobox, cursor), 
	BLT_CONFIG_NULL_OK},
    {BLT_CONFIG_DISTANCE, "-borderwidth", "borderWidth", "BorderWidth",
	DEF_COMBOBOX_BORDER_WIDTH, Blt_Offset(Combobox, borderWidth),
	BLT_CONFIG_DONT_SET_DEFAULT},
    {BLT_CONFIG_BOOLEAN, "-exportselection", "exportSelection",
	"ExportSelection", DEF_COMBOBOX_EXPORT_SELECTION, 
	Blt_Offset(Combobox, exportSelection), 
	BLT_CONFIG_DONT_SET_DEFAULT},
    {BLT_CONFIG_RELIEF, "-relief", "relief", "Relief",
	DEF_COMBOBOX_RELIEF, Blt_Offset(Combobox, relief), 0},
    {BLT_CONFIG_BORDER, "-selectbackground", "selectBackground", "Background",
	DEF_COMBOBOX_SELECT_BG_MONO, Blt_Offset(Combobox, selBorder),
	BLT_CONFIG_MONO_ONLY},
    {BLT_CONFIG_BORDER, "-selectbackground", "selectBackground", "Background",
	DEF_COMBOBOX_SELECT_BG_COLOR, Blt_Offset(Combobox, selBorder),
	BLT_CONFIG_COLOR_ONLY},
    {BLT_CONFIG_DISTANCE, "-selectborderwidth", "selectBorderWidth", 
        "BorderWidth", DEF_COMBOBOX_SELECT_BORDER_WIDTH, 
	Blt_Offset(Combobox, selBorderWidth), 
	BLT_CONFIG_DONT_SET_DEFAULT},
    {BLT_CONFIG_COLOR, "-selectforeground", "selectForeground", "Foreground",

	DEF_COMBOBOX_SELECT_FG_MONO, Blt_Offset(Combobox, selFgColor),
	BLT_CONFIG_MONO_ONLY},
    {BLT_CONFIG_COLOR, "-selectforeground", "selectForeground", "Foreground",
	DEF_COMBOBOX_SELECT_FG_COLOR, Blt_Offset(Combobox, selFgColor),
	BLT_CONFIG_COLOR_ONLY},
    {BLT_CONFIG_RELIEF, "-selectrelief", "selectRelief", "Relief",
	DEF_COMBOBOX_SELECT_RELIEF, Blt_Offset(Combobox, selRelief),
	BLT_CONFIG_DONT_SET_DEFAULT},
    {BLT_CONFIG_END, (char *)NULL, (char *)NULL, (char *)NULL, (char *)NULL, 
	0, 0}
};

static Tk_LostSelProc ComboboxLostSelectionProc;
static Tk_SelectionProc ComboboxSelectionProc;
static Tk_EventProc ComboboxEventProc;

/*
 *----------------------------------------------------------------------
 *
 * EventuallyRedraw --
 *
 *	Queues a request to redraw the widget at the next idle point.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Information gets redisplayed.  Right now we don't do selective
 *	redisplays:  the whole window will be redrawn.
 *
 *----------------------------------------------------------------------
 */
static void
EventuallyRedraw(comboPtr)
    Combobox *comboPtr;
{
    if ((comboPtr->tkwin != NULL) && 
	((comboPtr->flags & COMBOBOX_REDRAW) == 0)) {
	comboPtr->flags |= COMBOBOX_REDRAW;
	Tcl_DoWhenIdle(DisplayCombobox, comboPtr);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * BlinkCursorProc --
 *
 *	This procedure is called as a timer handler to blink the
 *	insertion cursor off and on.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The cursor gets turned on or off, redisplay gets invoked,
 *	and this procedure reschedules itself.
 *
 *----------------------------------------------------------------------
 */
static void
BlinkCursorProc(clientData)
    ClientData clientData;	/* Pointer to record describing entry. */
{
    Combobox *comboPtr = clientData;
    int interval;

    if (!(comboPtr->flags & COMBOBOX_FOCUS) || (comboPtr->offTime == 0)) {
	return;
    }
    if (comboPtr->active) {
	comboPtr->cursorOn ^= 1;
	interval = (comboPtr->cursorOn) ? comboPtr->onTime : comboPtr->offTime;
	comboPtr->timerToken = 
	    Tcl_CreateTimerHandler(interval, BlinkCursorProc, comboPtr);
	EventuallyRedraw(comboPtr);
    }
}

/*
 * --------------------------------------------------------------
 *
 * ComboboxEventProc --
 *
 * 	This procedure is invoked by the Tk dispatcher for various
 * 	events on hierarchy widgets.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 * --------------------------------------------------------------
 */
static void
ComboboxEventProc(clientData, eventPtr)
    ClientData clientData;	/* Information about window. */
    XEvent *eventPtr;		/* Information about event. */
{
    Combobox *comboPtr = clientData;

    if (eventPtr->type == Expose) {
	if (eventPtr->xexpose.count == 0) {
	    EventuallyRedraw(comboPtr);
	}
    } else if (eventPtr->type == ConfigureNotify) {
	EventuallyRedraw(comboPtr);
    } else if ((eventPtr->type == FocusIn) || (eventPtr->type == FocusOut)) {
	if (eventPtr->xfocus.detail == NotifyInferior) {
	    return;
	}
	if (eventPtr->type == FocusIn) {
	    comboPtr->flags |= COMBOBOX_FOCUS;
	} else {
	    comboPtr->flags &= ~COMBOBOX_FOCUS;
	}
	Tcl_DeleteTimerHandler(comboPtr->timerToken);
	if ((comboPtr->active) && (comboPtr->flags & COMBOBOX_FOCUS)) {
	    comboPtr->cursorOn = TRUE;
	    if (comboPtr->offTime != 0) {
		comboPtr->timerToken = Tcl_CreateTimerHandler(comboPtr->onTime, 
		   BlinkCursorProc, comboPtr);
	    }
	} else {
	    comboPtr->cursorOn = FALSE;
	    comboPtr->timerToken = (Tcl_TimerToken) NULL;
	}
	EventuallyRedraw(comboPtr);
    } else if (eventPtr->type == DestroyNotify) {
	if (comboPtr->tkwin != NULL) {
	    comboPtr->tkwin = NULL;
	}
	if (comboPtr->flags & COMBOBOX_REDRAW) {
	    Tcl_CancelIdleCall(DisplayCombobox, comboPtr);
	}
	if (comboPtr->timerToken != NULL) {
	    Tcl_DeleteTimerHandler(comboPtr->timerToken);
	}
	comboPtr->tvPtr->comboWin = NULL;
	Tcl_EventuallyFree(comboPtr, DestroyCombobox);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * ComboboxLostSelectionProc --
 *
 *	This procedure is called back by Tk when the selection is
 *	grabbed away from a Text widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The existing selection is unhighlighted, and the window is
 *	marked as not containing a selection.
 *
 *----------------------------------------------------------------------
 */
static void
ComboboxLostSelectionProc(clientData)
    ClientData clientData;	/* Information about Text widget. */
{
    Combobox *comboPtr = clientData;

    if ((comboPtr->selFirst >= 0) && (comboPtr->exportSelection)) {
	comboPtr->selFirst = comboPtr->selLast = -1;
	EventuallyRedraw(comboPtr);
    }
}

static int
PointerToIndex(comboPtr, x, y)
    Combobox *comboPtr;
    int x, y;
{
    TextLayout *textPtr;
    Tk_FontMetrics fontMetrics;
    TextFragment *fragPtr;
    int nBytes;
    register int i;
    int total;

    if ((comboPtr->string == NULL) || (comboPtr->string[0] == '\0')) {
	return 0;
    }
    x -= comboPtr->selBorderWidth;
    y -= comboPtr->selBorderWidth;

    textPtr = comboPtr->textPtr;

    /* Bound the y-coordinate within the window. */
    if (y < 0) {
	y = 0;
    } else if (y >= textPtr->height) {
	y = textPtr->height - 1;
    }
    /* 
     * Compute the line that contains the y-coordinate. 
     *
     * FIXME: This assumes that segments are distributed 
     *	     line-by-line.  This may change in the future.
     */
    Tk_GetFontMetrics(comboPtr->font, &fontMetrics);
    fragPtr = textPtr->fragArr;
    total = 0;
    for (i = (y / fontMetrics.linespace); i > 0; i--) {
	total += fragPtr->count;
	fragPtr++;
    }
    if (x < 0) {
	nBytes = 0;
    } else if (x >= textPtr->width) {
	nBytes = fragPtr->count;
    } else {
	int newX;

	/* Find the character underneath the pointer. */
	nBytes = Tk_MeasureChars(comboPtr->font, fragPtr->text, fragPtr->count, 
		 x, 0, &newX);
	if ((newX < x) && (nBytes < fragPtr->count)) {
	    double fract;
	    int length, charSize;
	    char *next;

	    next = fragPtr->text + nBytes;
#if HAVE_UTF
	    {
		Tcl_UniChar dummy;

		length = Tcl_UtfToUniChar(next, &dummy);
	    }
#else
	    length = 1;
#endif
	    charSize = Tk_TextWidth(comboPtr->font, next, length);
	    fract = ((double)(x - newX) / (double)charSize);
	    if (ROUND(fract)) {
		nBytes += length;
	    }
	}
    }
    return nBytes + total;
}

static int
IndexToPointer(comboPtr)
    Combobox *comboPtr;
{
    int x, y;
    int maxLines;
    TextLayout *textPtr;
    Tk_FontMetrics fontMetrics;
    int nBytes;
    int sum;
    TextFragment *fragPtr;
    register int i;

    textPtr = comboPtr->textPtr;
    Tk_GetFontMetrics(comboPtr->font, &fontMetrics);
    maxLines = (textPtr->height / fontMetrics.linespace) - 1;

    nBytes = sum = 0;
    x = y = comboPtr->borderWidth;
    if (comboPtr->icon != NULL) {
	x += TreeViewIconWidth(comboPtr->icon) + 2 * comboPtr->gap;
    }
    fragPtr = textPtr->fragArr;
    for (i = 0; i <= maxLines; i++) {
	/* Total the number of bytes on each line.  Include newlines. */
	nBytes = fragPtr->count + 1;
	if ((sum + nBytes) > comboPtr->insertPos) {
	    x += Tk_TextWidth(comboPtr->font, fragPtr->text, 
		comboPtr->insertPos - sum);
	    break;
	}
	y += fontMetrics.linespace;
	sum += nBytes;
	fragPtr++;
    }
    comboPtr->cursorX = x;
    comboPtr->cursorY = y;
    comboPtr->cursorHeight = fontMetrics.linespace;
    comboPtr->cursorWidth = 3;
    return TCL_OK;
}

static void
UpdateLayout(comboPtr)
    Combobox *comboPtr;
{
    TextStyle ts;
    int width, height;
    TextLayout *textPtr;
    int gap;
    int iconWidth, iconHeight;

    gap = iconWidth = iconHeight = 0;
    if (comboPtr->icon != NULL) {
	iconWidth = TreeViewIconWidth(comboPtr->icon) + 4;
	iconHeight = TreeViewIconHeight(comboPtr->icon);
	gap = comboPtr->gap;
    }

    /* The layout is based upon the current font. */
    Blt_InitTextStyle(&ts);
    ts.anchor = TK_ANCHOR_NW;
    ts.justify = TK_JUSTIFY_LEFT;
    ts.font = comboPtr->font;
    textPtr = Blt_GetTextLayout(comboPtr->string, &ts);
    if (comboPtr->textPtr != NULL) {
	Blt_Free(comboPtr->textPtr);
    }
    comboPtr->textPtr = textPtr;

    width = iconWidth + textPtr->width + gap * 2;
    height = MAX(iconHeight, textPtr->height);

    if (width <= comboPtr->columnPtr->width) {
	width = comboPtr->columnPtr->width;
    }
    if (height < comboPtr->entryPtr->height) {
	height = comboPtr->entryPtr->height;
    }
    comboPtr->width = width + (2 * comboPtr->borderWidth);
    comboPtr->height = height + (2 * comboPtr->borderWidth);
    IndexToPointer(comboPtr);
    Blt_MoveResizeToplevel(comboPtr->tkwin, comboPtr->x, comboPtr->y, 
	      comboPtr->width, comboPtr->height);
    Blt_MapToplevel(comboPtr->tkwin);
    Blt_RaiseToplevel(comboPtr->tkwin);
}

static void
InsertText(comboPtr, insertText, insertPos, nBytes)
    Combobox *comboPtr;
    char *insertText;
    int insertPos;
    int nBytes;
{
    int oldSize, newSize;
    char *oldText, *newText;

    oldText = comboPtr->string;
    oldSize = strlen(oldText);
    newSize = oldSize + nBytes;
    newText = Blt_Malloc(sizeof(char) * (newSize + 1));
    if (insertPos == oldSize) {	/* Append */
	strcpy(newText, oldText);
	strcat(newText, insertText);
    } else if (insertPos == 0) {/* Prepend */
	strcpy(newText, insertText);
	strcat(newText, oldText);
    } else {			/* Insert into existing. */
	char *p;
	
	p = newText;
	strncpy(p, oldText, insertPos);
	p += insertPos;
	strcpy(p, insertText);
	p += nBytes;
	strcpy(p, oldText + insertPos);
    }

    /* 
     * All indices from the start of the insertion to the end of the
     * string need to be updated.  Simply move the indices down by the
     * number of characters added.  
     */
    if (comboPtr->selFirst >= insertPos) {
	comboPtr->selFirst += nBytes;
    }
    if (comboPtr->selLast > insertPos) {
	comboPtr->selLast += nBytes;
    }
    if ((comboPtr->selAnchor > insertPos) || (comboPtr->selFirst >= insertPos)) {
	comboPtr->selAnchor += nBytes;
    }
    if (comboPtr->string != NULL) {
	Blt_Free(comboPtr->string);
    }
    comboPtr->string = newText;
    comboPtr->insertPos = insertPos + nBytes;
    UpdateLayout(comboPtr);
}

static int
DeleteText(comboPtr, firstPos, lastPos)
    Combobox *comboPtr;
    int firstPos, lastPos;
{
    char *oldText, *newText;
    int oldSize, newSize;
    int nBytes;
    char *p;

    oldText = comboPtr->string;
    if (firstPos > lastPos) {
	return TCL_OK;
    }
    lastPos++;			/* Now is the position after the last
				 * character. */

    nBytes = lastPos - firstPos;

    oldSize = strlen(oldText) + 1;
    newSize = oldSize - nBytes + 1;
    newText = Blt_Malloc(sizeof(char) * newSize);
    p = newText;
    if (firstPos > 0) {
	strncpy(p, oldText, firstPos);
	p += firstPos;
    }
    *p = '\0';
    if (lastPos < oldSize) {
	strcpy(p, oldText + lastPos);
    }
    Blt_Free(oldText);

    /*
     * Since deleting characters compacts the character array, we need to
     * update the various character indices according.  It depends where
     * the index occurs in relation to range of deleted characters:
     *
     *	 before		Ignore.
     *   within		Move the index back to the start of the deletion.
     *	 after		Subtract off the deleted number of characters,
     *			since the array has been compressed by that
     *			many characters.
     *
     */
    if (comboPtr->selFirst >= firstPos) {
	if (comboPtr->selFirst >= lastPos) {
	    comboPtr->selFirst -= nBytes;
	} else {
	    comboPtr->selFirst = firstPos;
	}
    }
    if (comboPtr->selLast >= firstPos) {
	if (comboPtr->selLast >= lastPos) {
	    comboPtr->selLast -= nBytes;
	} else {
	    comboPtr->selLast = firstPos;
	}
    }
    if (comboPtr->selLast <= comboPtr->selFirst) {
	comboPtr->selFirst = comboPtr->selLast = -1; /* Cut away the entire
						    * selection. */ 
    }
    if (comboPtr->selAnchor >= firstPos) {
	if (comboPtr->selAnchor >= lastPos) {
	    comboPtr->selAnchor -= nBytes;
	} else {
	    comboPtr->selAnchor = firstPos;
	}
    }
    if (comboPtr->insertPos >= firstPos) {
	if (comboPtr->insertPos >= lastPos) {
	    comboPtr->insertPos -= nBytes;
	} else {
	    comboPtr->insertPos = firstPos;
	}
    }
    comboPtr->string = newText;
    UpdateLayout(comboPtr);
    EventuallyRedraw(comboPtr);
    return TCL_OK;
}

static int
AcquireText(tvPtr, comboPtr, entryPtr, columnPtr)
    TreeView *tvPtr;
    Combobox *comboPtr;
    TreeViewEntry *entryPtr;
    TreeViewColumn *columnPtr;
{
    TreeViewStyle *stylePtr;
    int x, y;
    char *string;
    TreeViewIcon icon;

    if (columnPtr == &tvPtr->treeColumn) {
	int level;

	level = DEPTH(tvPtr, entryPtr->node);
	x = SCREENX(tvPtr, entryPtr->worldX);
	y = SCREENY(tvPtr, entryPtr->worldY);
	x += ICONWIDTH(level) + ICONWIDTH(level + 1) + 4;
	string = GETLABEL(entryPtr);
	stylePtr = columnPtr->stylePtr;
	icon = Blt_TreeViewGetEntryIcon(tvPtr, entryPtr);
    } else {
	TreeViewValue *valuePtr;

	x = SCREENX(tvPtr, columnPtr->worldX);
	y = SCREENY(tvPtr, entryPtr->worldY);
	stylePtr = columnPtr->stylePtr;
	valuePtr = Blt_TreeViewFindValue(entryPtr, columnPtr);
	string = valuePtr->string;
	if (valuePtr->stylePtr != NULL) {
	    stylePtr = valuePtr->stylePtr;
	}
	icon = stylePtr->icon;
    }
    if (comboPtr->textPtr != NULL) {
	Blt_Free(comboPtr->textPtr);
	comboPtr->textPtr = NULL;
    }
    if (comboPtr->string != NULL) {
	Blt_Free(comboPtr->string);
    }
    if (string == NULL) {
	string = "";
    }
    /* Adjust current location for virtual root windows.  */
    comboPtr->icon = icon;
    comboPtr->entryPtr = entryPtr;
    comboPtr->columnPtr = columnPtr;
    Blt_RootCoordinates(tvPtr->tkwin, x - comboPtr->borderWidth,
	y - comboPtr->borderWidth, &comboPtr->x, &comboPtr->y);
    comboPtr->gap = stylePtr->gap;
    comboPtr->string = Blt_Strdup(string);
    comboPtr->gc = Blt_TreeViewGetStyleGC(tvPtr, stylePtr);
    comboPtr->font = Blt_TreeViewGetStyleFont(tvPtr, stylePtr);
    comboPtr->selFirst = comboPtr->selLast = -1;
    UpdateLayout(comboPtr);
    Blt_MapToplevel(comboPtr->tkwin);
    Tk_MapWindow(comboPtr->tkwin);
    EventuallyRedraw(comboPtr);
    return TCL_OK;
}


/*
 *---------------------------------------------------------------------------
 *
 * GetIndexFromObj --
 *
 *	Parse an index into an entry and return either its value
 *	or an error.
 *
 * Results:
 *	A standard Tcl result.  If all went well, then *indexPtr is
 *	filled in with the character index (into entryPtr) corresponding to
 *	string.  The index value is guaranteed to lie between 0 and
 *	the number of characters in the string, inclusive.  If an
 *	error occurs then an error message is left in the interp's result.
 *
 * Side effects:
 *	None.
 *
 *---------------------------------------------------------------------------
 */
static int
GetIndexFromObj(interp, comboPtr, objPtr, indexPtr)
    Tcl_Interp *interp;
    Combobox *comboPtr;
    Tcl_Obj *objPtr;
    int *indexPtr;
{
    int textPos;
    char c;
    char *string;

    string = Tcl_GetString(objPtr);
    if ((comboPtr->string == NULL) || (comboPtr->string[0] == '\0')) {
	*indexPtr = 0;
	return TCL_OK;
    }
    c = string[0];
    if ((c == 'a') && (strcmp(string, "anchor") == 0)) {
	textPos = comboPtr->selAnchor;
    } else if ((c == 'e') && (strcmp(string, "end") == 0)) {
	textPos = strlen(comboPtr->string);
    } else if ((c == 'i') && (strcmp(string, "insert") == 0)) {
	textPos = comboPtr->insertPos;
    } else if ((c == 'n') && (strcmp(string, "next") == 0)) {
	textPos = comboPtr->insertPos;
	if (textPos < (int)strlen(comboPtr->string)) {
	    textPos++;
	}
    } else if ((c == 'l') && (strcmp(string, "last") == 0)) {
	textPos = comboPtr->insertPos;
	if (textPos > 0) {
	    textPos--;
	}
    } else if ((c == 's') && (strcmp(string, "sel.first") == 0)) {
	if (comboPtr->selFirst < 0) {
	    textPos = -1;
	} else {
	    textPos = comboPtr->selFirst;
	}
    } else if ((c == 's') && (strcmp(string, "sel.last") == 0)) {
	if (comboPtr->selLast < 0) {
	    textPos = -1;
	} else {
	    textPos = comboPtr->selLast;
	}
    } else if (c == '@') {
	int x, y;

	if (Blt_GetXY(interp, comboPtr->tkwin, string, &x, &y) != TCL_OK) {
	    return TCL_ERROR;
	}
	textPos = PointerToIndex(comboPtr, x, y);
    } else if (isdigit((int)c)) {
	int number;
	int maxChars;

	if (Tcl_GetIntFromObj(interp, objPtr, &number) != TCL_OK) {
	    return TCL_ERROR;
	}
	/* Don't allow the index to point outside the label. */
	maxChars = Tcl_NumUtfChars(comboPtr->string, -1);
	if (number < 0) {
	    textPos = 0;
	} else if (number > maxChars) {
	    textPos = strlen(comboPtr->string);
	} else {
	    textPos = Tcl_UtfAtIndex(comboPtr->string, number) - 
		comboPtr->string;
	}
    } else {
	if (interp != NULL) {
	    Tcl_AppendResult(interp, "bad label index \"", string, "\"",
			     (char *)NULL);
	}
	return TCL_ERROR;
    }
    *indexPtr = textPos;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * SelectText --
 *
 *	Modify the selection by moving its un-anchored end.  This
 *	could make the selection either larger or smaller.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The selection changes.
 *
 *----------------------------------------------------------------------
 */
static int
SelectText(comboPtr, textPos)
    Combobox *comboPtr; 	/* Information about combobox. */
    int textPos;		/* Index of element that is to
				 * become the "other" end of the
				 * selection. */
{
    int selFirst, selLast;

    /*
     * Grab the selection if we don't own it already.
     */
    if ((comboPtr->exportSelection) && (comboPtr->selFirst == -1)) {
	Tk_OwnSelection(comboPtr->tkwin, XA_PRIMARY, 
			ComboboxLostSelectionProc, comboPtr);
    }
    /*  If the anchor hasn't been set yet, assume the beginning of the text*/
    if (comboPtr->selAnchor < 0) {
	comboPtr->selAnchor = 0;
    }
    if (comboPtr->selAnchor <= textPos) {
	selFirst = comboPtr->selAnchor;
	selLast = textPos;
    } else {
	selFirst = textPos;
	selLast = comboPtr->selAnchor;
    }
    if ((comboPtr->selFirst != selFirst) || (comboPtr->selLast != selLast)) {
	comboPtr->selFirst = selFirst;
	comboPtr->selLast = selLast;
	EventuallyRedraw(comboPtr);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ComboboxSelectionProc --
 *
 *	This procedure is called back by Tk when the selection is
 *	requested by someone.  It returns part or all of the selection
 *	in a buffer provided by the caller.
 *
 * Results:
 *	The return value is the number of non-NULL bytes stored at
 *	buffer.  Buffer is filled (or partially filled) with a
 *	NUL-terminated string containing part or all of the
 *	selection, as given by offset and maxBytes.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */
static int
ComboboxSelectionProc(clientData, offset, buffer, maxBytes)
    ClientData clientData;	/* Information about the widget. */
    int offset;			/* Offset within selection of first
				 * character to be returned. */
    char *buffer;		/* Location in which to place
				 * selection. */
    int maxBytes;		/* Maximum number of bytes to place
				 * at buffer, not including terminating
				 * NULL character. */
{
    Combobox *comboPtr = clientData;
    int size;

    size = strlen(comboPtr->string + offset);
    /*
     * Return the string currently in the combobox.
     */
    strncpy(buffer, comboPtr->string + offset, maxBytes);
    buffer[maxBytes] = '\0';
    return (size > maxBytes) ? maxBytes : size;
}


static void
DestroyCombobox(data)
    DestroyData data;
{
    Combobox *comboPtr = (Combobox *)data;

    Blt_FreeObjOptions(comboboxConfigSpecs, (char *)comboPtr, comboPtr->display, 0);

    if (comboPtr->string != NULL) {
	Blt_Free(comboPtr->string);
    }
    if (comboPtr->textPtr != NULL) {
	Blt_Free(comboPtr->textPtr);
    }
    if (comboPtr->timerToken != NULL) {
	Tcl_DeleteTimerHandler(comboPtr->timerToken);
    }
    if (comboPtr->tkwin != NULL) {
	Tk_DeleteSelHandler(comboPtr->tkwin, XA_PRIMARY, XA_STRING);
    }
    Blt_Free(comboPtr);
}

static void
ConfigureCombobox(comboPtr)
    Combobox *comboPtr;
{
#ifdef notdef
    GC newGC;
    Combobox *comboPtr = tvPtr->comboPtr;
    XGCValues gcValues;
    unsigned long gcMask;

    /*
     * GC for edit window.
     */
    gcMask = 0;
    newGC = Tk_GetGC(comboPtr->tkwin, gcMask, &gcValues);
    if (comboPtr->gc != NULL) {
	Tk_FreeGC(comboPtr->display, comboPtr->gc);
    }
    comboPtr->gc = newGC;
    comboPtr->width = comboPtr->textPtr->width + 
	2 * (comboPtr->borderWidth + comboPtr->highlightWidth);
    comboPtr->height = comboPtr->textPtr->height + 
	2 * (comboPtr->borderWidth + comboPtr->highlightWidth);
    
    if (Tk_IsMapped(comboPtr->tkwin)) {
	if ((comboPtr->height != Tk_Height(comboPtr->tkwin)) ||
	    (comboPtr->width != Tk_Width(comboPtr->tkwin))) {
	    Tk_ResizeWindow(comboPtr->tkwin, comboPtr->width, comboPtr->height);
	}
    }
#endif
}

void
Blt_TreeViewCombobox(tvPtr, entryPtr, columnPtr)
    TreeView *tvPtr;
    TreeViewEntry *entryPtr;
    TreeViewColumn *columnPtr;
{
    Tk_Window tkwin;
    Combobox *comboPtr;
    XSetWindowAttributes attrs;
    char editClass[20];
    unsigned int mask;

    tkwin = Tk_CreateWindow(tvPtr->interp, tvPtr->tkwin, "edit", "");
    if (tkwin == NULL) {
	return;
    }

    attrs.override_redirect = True;
    attrs.backing_store = WhenMapped;
    attrs.save_under = True;
    mask = CWOverrideRedirect | CWSaveUnder | CWBackingStore;
    Tk_ChangeWindowAttributes(tkwin, mask, &attrs);
    Tk_MakeWindowExist(tkwin);

    sprintf(editClass, "%sEditor", Tk_Class(tvPtr->tkwin));
    Tk_SetClass(tkwin, editClass); 

    comboPtr = Blt_Calloc(1, sizeof(Combobox));
    assert(comboPtr);

    comboPtr->interp = tvPtr->interp;
    comboPtr->display = Tk_Display(tkwin);
    comboPtr->tkwin = tkwin;
    comboPtr->borderWidth = 1;
    comboPtr->relief = TK_RELIEF_SOLID;
    comboPtr->selRelief = TK_RELIEF_FLAT;
    comboPtr->selBorderWidth = 1;
    comboPtr->selAnchor = -1;
    comboPtr->selFirst = comboPtr->selLast = -1;
    comboPtr->onTime = 600;
    comboPtr->active = TRUE;
    comboPtr->offTime = 300;
    comboPtr->tvPtr = tvPtr;
    tvPtr->comboWin = tkwin;
#if (TK_MAJOR_VERSION > 4)
    Blt_SetWindowInstanceData(tkwin, comboPtr);
#endif
    Tk_CreateSelHandler(tkwin, XA_PRIMARY, XA_STRING, 
	ComboboxSelectionProc, comboPtr, XA_STRING);
    Tk_CreateEventHandler(tkwin, ExposureMask | StructureNotifyMask |
	FocusChangeMask, ComboboxEventProc, comboPtr);
    Tcl_CreateObjCommand(tvPtr->interp, Tk_PathName(tkwin), 
	ComboboxCmd, comboPtr, NULL);
    if (Blt_ConfigureWidgetFromObj(tvPtr->interp, tkwin, comboboxConfigSpecs, 0, 
	(Tcl_Obj **)NULL, (char *)comboPtr, 0) != TCL_OK) {
	Tk_DestroyWindow(tkwin);
	return;
    }
    AcquireText(tvPtr, comboPtr, entryPtr, columnPtr);
    comboPtr->insertPos = strlen(comboPtr->string);
    
    Blt_MoveResizeToplevel(tkwin, comboPtr->x, comboPtr->y, comboPtr->width, 
	comboPtr->height);
    Blt_MapToplevel(tkwin);
    Blt_RaiseToplevel(tkwin);
    Tk_MapWindow(tkwin);
    EventuallyRedraw(comboPtr);
}

static void
DisplayCombobox(clientData)
    ClientData clientData;
{
    Combobox *comboPtr = clientData;
    Pixmap drawable;
    register int i;
    int x1, x2;
    int count, nChars;
    int leftPos, rightPos;
    int selStart, selEnd, selLength;
    int x, y;
    TextFragment *fragPtr;
    Tk_FontMetrics fontMetrics;

    comboPtr->flags &= ~COMBOBOX_REDRAW;
    if (!Tk_IsMapped(comboPtr->tkwin)) {
	return;
    }
    if (comboPtr->columnPtr == NULL) {
	return;
    }
    drawable = Tk_GetPixmap(comboPtr->display, Tk_WindowId(comboPtr->tkwin), 
	Tk_Width(comboPtr->tkwin), Tk_Height(comboPtr->tkwin), 
	Tk_Depth(comboPtr->tkwin));

    Tk_Fill3DRectangle(comboPtr->tkwin, drawable, comboPtr->border, 0, 0,
	Tk_Width(comboPtr->tkwin), Tk_Height(comboPtr->tkwin), 
	comboPtr->borderWidth, comboPtr->relief);

    x = comboPtr->borderWidth + comboPtr->gap;
    y = comboPtr->borderWidth;
    if (comboPtr->icon != NULL) {
	y += (comboPtr->height - TreeViewIconHeight(comboPtr->icon)) / 2;
	Tk_RedrawImage(TreeViewIconBits(comboPtr->icon), 0, 0, 
		       TreeViewIconWidth(comboPtr->icon), 
		       TreeViewIconHeight(comboPtr->icon), 
		       drawable, x, y);
	x += TreeViewIconWidth(comboPtr->icon) + comboPtr->gap;
    }
    
    Tk_GetFontMetrics(comboPtr->font, &fontMetrics);
    fragPtr = comboPtr->textPtr->fragArr;
    count = 0;
    y = comboPtr->borderWidth;
    if (comboPtr->height > fontMetrics.linespace) {
	y += (comboPtr->height - fontMetrics.linespace) / 2;
    }
    for (i = 0; i < comboPtr->textPtr->nFrags; i++, fragPtr++) {
	leftPos = count;
	count += fragPtr->count;
	rightPos = count;
	if ((rightPos < comboPtr->selFirst) || (leftPos > comboPtr->selLast)) {
	    /* No part of the text fragment is selected. */
	    Tk_DrawChars(comboPtr->display, drawable, comboPtr->gc, comboPtr->font,
			 fragPtr->text, fragPtr->count, 
			 x + fragPtr->x, y + fragPtr->y);
	    continue;
	}

	/*
	 *  A text fragment can have up to 3 regions:
	 *
	 *	1. Text before the start the selection
	 *	2. Selected text itself (drawn in a raised border)
	 *	3. Text following the selection.
	 */

	selStart = leftPos;
	selEnd = rightPos;
	/* First adjust selected region for current line. */
	if (comboPtr->selFirst > leftPos) {
	    selStart = comboPtr->selFirst;
	}
	if (comboPtr->selLast < rightPos) {
	    selEnd = comboPtr->selLast;
	}
	selLength = (selEnd - selStart);
	x1 = x;

	if (selStart > leftPos) { /* Normal text preceding the selection */
	    nChars = (selStart - leftPos);
	    Tk_MeasureChars(comboPtr->font, comboPtr->string + leftPos,
		    nChars, 10000, DEF_TEXT_FLAGS, &x1);
	    x1 += x;
	}
	if (selLength > 0) {	/* The selection itself */
	    int width;

	    Tk_MeasureChars(comboPtr->font, fragPtr->text + selStart, selLength,
		10000, DEF_TEXT_FLAGS, &x2);
	    x2 += x;
	    width = (x2 - x1) + 1;
	    Tk_Fill3DRectangle(comboPtr->tkwin, drawable, comboPtr->selBorder,
		x1, y + fragPtr->y - fontMetrics.ascent, 
	        width, fontMetrics.linespace,
		comboPtr->selBorderWidth, comboPtr->selRelief);
	}
	Tk_DrawChars(Tk_Display(comboPtr->tkwin), drawable, comboPtr->gc, 
	     comboPtr->font, fragPtr->text, fragPtr->count, 
	     fragPtr->x + x, fragPtr->y + y);
    }
    if ((comboPtr->flags & COMBOBOX_FOCUS) && (comboPtr->cursorOn)) {
	int left, top, right, bottom;

	IndexToPointer(comboPtr);
	left = comboPtr->cursorX + 1;
	right = left + 1;
	top = comboPtr->cursorY;
	if (comboPtr->height > fontMetrics.linespace) {
	    top += (comboPtr->height - fontMetrics.linespace) / 2;
	}
	bottom = top + comboPtr->cursorHeight - 1;
	XDrawLine(comboPtr->display, drawable, comboPtr->gc, left, top, left,
		bottom);
	XDrawLine(comboPtr->display, drawable, comboPtr->gc, left - 1, top, right,
		top);
	XDrawLine(comboPtr->display, drawable, comboPtr->gc, left - 1, bottom, 
		right, bottom);
    }
#ifdef notdef
    {
	int buttonX, buttonY, buttonWidth, buttonHeight;

	buttonWidth = SORT_MARKER_WIDTH + 6;
	buttonHeight = Tk_Height(comboPtr->tkwin) - 4;
	Tk_Fill3DRectangle(combotr->tkwin, drawable, comboPtr->border, 
	   buttonX, buttonY + OB_BORDERWIDTH, buttonWidth, buttonHeight,
	   comboPtr->buttonBorderWidth, comboPtr->buttonRelief); 

	buttonX = Tk_Width(comboPtr->tkwin) - 
	    comboPtr->buttonWidth / 2 - (ARROW_OFFSET + 1);
	buttonY = comboPtr->height / 2;
	triangle[0].x = buttonX + ARROW_OFFSET + 1;
	triangle[0].y = buttonY + ARROW_OFFSET / 2;
	triangle[1].x = triangle[0].x - ARROW_OFFSET;
	triangle[1].y = triangle[0].y - ARROW_OFFSET;
	triangle[2].x = triangle[0].x + ARROW_OFFSET;
	triangle[2].y = triangle[0].y - ARROW_OFFSET;
	triangle[3].x = triangle[0].x;
	triangle[3].y = triangle[0].y;

	XFillPolygon(comboPtr->display, drawable, gc, triangle, 4, Convex, 
		     CoordModeOrigin);
	XDrawLines(comboPtr->display, drawable, gc, triangle, 4, 
		   CoordModeOrigin);
    }
#endif
    Tk_Draw3DRectangle(comboPtr->tkwin, drawable, comboPtr->border, 0, 0,
	Tk_Width(comboPtr->tkwin), Tk_Height(comboPtr->tkwin), 
	comboPtr->borderWidth, comboPtr->relief);
    XCopyArea(comboPtr->display, drawable, Tk_WindowId(comboPtr->tkwin),
	comboPtr->gc, 0, 0, Tk_Width(comboPtr->tkwin), 
	Tk_Height(comboPtr->tkwin), 0, 0);
    Tk_FreePixmap(comboPtr->display, drawable);

}

/*ARGSUSED*/
static int
ApplyOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		
    int objc;			/* Not used. */
    Tcl_Obj *CONST *objv;	/* Not used. */
{
    TreeViewEntry *entryPtr;
    TreeView *tvPtr = comboPtr->tvPtr;

    entryPtr = comboPtr->entryPtr;
    if (comboPtr->columnPtr == &tvPtr->treeColumn) {
	if (entryPtr->labelUid != NULL) {
	    Blt_TreeViewFreeUid(tvPtr, entryPtr->labelUid);
	}
	if (comboPtr->string == NULL) {
	    entryPtr->labelUid = Blt_TreeViewGetUid(tvPtr, "");
	} else {
	    entryPtr->labelUid = Blt_TreeViewGetUid(tvPtr, comboPtr->string);
	}
    } else {
	TreeViewColumn *columnPtr;
	Tcl_Obj *objPtr;

	columnPtr = comboPtr->columnPtr;
	objPtr = Tcl_NewStringObj(comboPtr->string, -1);
	if (Blt_TreeSetValueByKey(interp, tvPtr->tree, entryPtr->node, 
		columnPtr->key, objPtr) != TCL_OK) {
	    Tcl_DecrRefCount(objPtr);
	    return TCL_ERROR;
	}
	entryPtr->flags |= ENTRY_DIRTY;
    }	
    if (tvPtr != NULL) {
	Blt_TreeViewConfigureEntry(tvPtr, entryPtr);
	tvPtr->flags |= (TV_LAYOUT | TV_DIRTY);
	Blt_TreeViewEventuallyRedraw(tvPtr);
    }
    Tk_DestroyWindow(comboPtr->tkwin);
    return TCL_OK;
}

/*ARGSUSED*/
static int
CancelOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		/* Not used. */
    int objc;			/* Not used. */
    Tcl_Obj *CONST *objv;	/* Not used. */
{
    Tk_DestroyWindow(comboPtr->tkwin);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CgetOp --
 *
 *----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
CgetOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;
    int objc;			/* Not used. */
    Tcl_Obj *CONST *objv;
{
    return Blt_ConfigureValueFromObj(interp, comboPtr->tkwin, 
	comboboxConfigSpecs, (char *)comboPtr, objv[2], 0);
}

/*
 *----------------------------------------------------------------------
 *
 * ConfigureOp --
 *
 * 	This procedure is called to process a list of configuration
 *	options database, in order to reconfigure the one of more
 *	entries in the widget.
 *
 *	  .c configure option value
 *
 * Results:
 *	A standard Tcl result.  If TCL_ERROR is returned, then
 *	interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as text string, colors, font,
 *	etc. get set for tvPtr; old resources get freed, if there
 *	were any.  The hypertext is redisplayed.
 *
 *----------------------------------------------------------------------
 */
static int
ConfigureOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST *objv;
{
    if (objc == 2) {
	return Blt_ConfigureInfoFromObj(interp, comboPtr->tkwin, 
		comboboxConfigSpecs, (char *)comboPtr, (Tcl_Obj *)NULL, 0);
    } else if (objc == 3) {
	return Blt_ConfigureInfoFromObj(interp, comboPtr->tkwin, 
		comboboxConfigSpecs, (char *)comboPtr, objv[3], 0);
    }
    if (Blt_ConfigureWidgetFromObj(interp, comboPtr->tkwin, 
	comboboxConfigSpecs, objc - 2, objv + 2, (char *)comboPtr, 
	BLT_CONFIG_OBJV_ONLY) != TCL_OK) {
	return TCL_ERROR;
    }
    ConfigureCombobox(comboPtr);
    EventuallyRedraw(comboPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DeleteOp --
 *
 *	Remove one or more characters from the label of an entry.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory gets freed, the entry gets modified and (eventually)
 *	redisplayed.
 *
 *----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
DeleteOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		/* Not used. */
    int objc;
    Tcl_Obj *CONST *objv;
{
    int firstPos, lastPos;

    if (comboPtr->entryPtr == NULL) {
	return TCL_OK;
    }
    if (GetIndexFromObj(interp, comboPtr, objv[2], &firstPos) != TCL_OK) {
	return TCL_ERROR;
    }
    lastPos = firstPos;
    if ((objc == 4) && 
	(GetIndexFromObj(interp, comboPtr, objv[3], &lastPos) != TCL_OK)) {
	return TCL_ERROR;
    }
    if (firstPos > lastPos) {
	return TCL_OK;
    }
    return DeleteText(comboPtr, firstPos, lastPos);
}


/*
 *----------------------------------------------------------------------
 *
 * IcursorOp --
 *
 *	Returns the numeric index of the given string. Indices can be
 *	one of the following:
 *
 *	"anchor"	Selection anchor.
 *	"end"		End of the label.
 *	"insert"	Insertion cursor.
 *	"sel.first"	First character selected.
 *	"sel.last"	Last character selected.
 *	@x,y		Index at X-Y screen coordinate.
 *	number		Returns the same number.
 *
 * Results:
 *	A standard Tcl result.  If the argument does not represent a
 *	valid label index, then TCL_ERROR is returned and the interpreter
 *	result will contain an error message.
 *
 *----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
IcursorOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;
    int objc;			/* Not used. */
    Tcl_Obj *CONST *objv;
{
    int textPos;

    if (GetIndexFromObj(interp, comboPtr, objv[2], &textPos) != TCL_OK) {
	return TCL_ERROR;
    }
    if (comboPtr->columnPtr != NULL) {
	comboPtr->insertPos = textPos;
	IndexToPointer(comboPtr);
	EventuallyRedraw(comboPtr);
    }
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * IndexOp --
 *
 *	Returns the numeric index of the given string. Indices can be
 *	one of the following:
 *
 *	"anchor"	Selection anchor.
 *	"end"		End of the label.
 *	"insert"	Insertion cursor.
 *	"sel.first"	First character selected.
 *	"sel.last"	Last character selected.
 *	@x,y		Index at X-Y screen coordinate.
 *	number		Returns the same number.
 *
 * Results:
 *	A standard Tcl result.  If the argument does not represent a
 *	valid label index, then TCL_ERROR is returned and the interpreter
 *	result will contain an error message.
 *
 *----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
IndexOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;
    int objc;			/* Not used. */
    Tcl_Obj *CONST *objv;
{
    int textPos;

    if (GetIndexFromObj(interp, comboPtr, objv[2], &textPos) != TCL_OK) {
	return TCL_ERROR;
    }
    if ((comboPtr->columnPtr != NULL) && (comboPtr->string != NULL)) {
	int nChars;

	nChars = Tcl_NumUtfChars(comboPtr->string, textPos);
	Tcl_SetObjResult(interp, Tcl_NewIntObj(nChars));
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * InsertOp --
 *
 *	Add new characters to the label of an entry.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	New information gets added to comboPtr;  it will be redisplayed
 *	soon, but not necessarily immediately.
 *
 *----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
InsertOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		/* Not used. */
    int objc;
    Tcl_Obj *CONST *objv;
{
    int extra;
    int insertPos;
    char *string;

    if (comboPtr->entryPtr == NULL) {
	return TCL_ERROR;
    }
    if (GetIndexFromObj(interp, comboPtr, objv[2], &insertPos) != TCL_OK) {
	return TCL_ERROR;
    }
    string = Tcl_GetString(objv[3]);
    extra = strlen(string);
    if (extra == 0) {	/* Nothing to insert. Move the cursor anyways. */
	comboPtr->insertPos = insertPos;
    } else {
	InsertText(comboPtr, string, insertPos, extra);
    }
    return TCL_OK;
}

/*ARGSUSED*/
static int
SelectionAdjustOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		/* Not used. */
    int objc;
    Tcl_Obj *CONST *objv;
{
    int textPos;
    int half1, half2;

    if (GetIndexFromObj(interp, comboPtr, objv[3], &textPos) != TCL_OK) {
	return TCL_ERROR;
    }
    half1 = (comboPtr->selFirst + comboPtr->selLast) / 2;
    half2 = (comboPtr->selFirst + comboPtr->selLast + 1) / 2;
    if (textPos < half1) {
	comboPtr->selAnchor = comboPtr->selLast;
    } else if (textPos > half2) {
	comboPtr->selAnchor = comboPtr->selFirst;
    }
    return SelectText(comboPtr, textPos);
}

/*ARGSUSED*/
static int
SelectionClearOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		/* Not used. */
    int objc;			/* Not used. */
    Tcl_Obj *CONST *objv;	/* Not used. */
{
    if (comboPtr->selFirst != -1) {
	comboPtr->selFirst = comboPtr->selLast = -1;
	EventuallyRedraw(comboPtr);
    }
    return TCL_OK;
}

/*ARGSUSED*/
static int
SelectionFromOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		/* Not used. */
    int objc;
    Tcl_Obj *CONST *objv;
{
    int textPos;

    if (GetIndexFromObj(interp, comboPtr, objv[3], &textPos) != TCL_OK) {
	return TCL_ERROR;
    }
    comboPtr->selAnchor = textPos;
    return TCL_OK;
}

/*ARGSUSED*/
static int
SelectionPresentOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		
    int objc;			/* Not used. */
    Tcl_Obj *CONST *objv;	/* Not used. */
{
    int bool;

    bool = (comboPtr->selFirst != -1);
    Tcl_SetObjResult(interp, Tcl_NewBooleanObj(bool));
    return TCL_OK;
}

/*ARGSUSED*/
static int
SelectionRangeOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		/* Not used. */
    int objc;
    Tcl_Obj *CONST *objv;
{
    int selFirst, selLast;

    if (GetIndexFromObj(interp, comboPtr, objv[3], &selFirst) != TCL_OK) {
	return TCL_ERROR;
    }
    if (GetIndexFromObj(interp, comboPtr, objv[4], &selLast) != TCL_OK) {
	return TCL_ERROR;
    }
    comboPtr->selAnchor = selFirst;
    return SelectText(comboPtr, selLast);
}

/*ARGSUSED*/
static int
SelectionToOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;		/* Not used. */
    int objc;
    Tcl_Obj *CONST *objv;
{
    int textPos;

    if (GetIndexFromObj(interp, comboPtr, objv[3], &textPos) != TCL_OK) {
	return TCL_ERROR;
    }
    return SelectText(comboPtr, textPos);
}


static Blt_OpSpec selectionOps[] =
{
    {"adjust", 1, (Blt_Op)SelectionAdjustOp, 4, 4, "index",},
    {"clear", 1, (Blt_Op)SelectionClearOp, 3, 3, "",},
    {"from", 1, (Blt_Op)SelectionFromOp, 4, 4, "index"},
    {"present", 1, (Blt_Op)SelectionPresentOp, 3, 3, ""},
    {"range", 1, (Blt_Op)SelectionRangeOp, 5, 5, "start end",},
    {"to", 1, (Blt_Op)SelectionToOp, 4, 4, "index"},
};

static int nSelectionOps = sizeof(selectionOps) / sizeof(Blt_OpSpec);

/*
 *	This procedure handles the individual options for text
 *	selections.  The selected text is designated by start and end
 *	indices into the text pool.  The selected segment has both a
 *	anchored and unanchored ends.  The following selection
 *	operations are implemented:
 *
 *	  "adjust"	- resets either the first or last index
 *			  of the selection.
 *	  "clear"	- clears the selection. Sets first/last
 *			  indices to -1.
 *	  "from"	- sets the index of the selection anchor.
 *	  "present"	- return "1" if a selection is available,
 *			  "0" otherwise.
 *	  "range"	- sets the first and last indices.
 *	  "to"		- sets the index of the un-anchored end.
 */
static int
SelectionOp(comboPtr, interp, objc, objv)
    Combobox *comboPtr;
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST *objv;
{
    Blt_Op proc;
    int result;

    proc = Blt_GetOpFromObj(interp, nSelectionOps, selectionOps, BLT_OP_ARG2, 
	objc, objv, 0);
    if (proc == NULL) {
	return TCL_ERROR;
    }
    result = (*proc) (comboPtr, interp, objc, objv);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * ComboboxCmd --
 *
 *	This procedure handles entry operations.
 *
 * Results:
 *	A standard Tcl result.
 *
 *----------------------------------------------------------------------
 */
static Blt_OpSpec comboOps[] =
{
    {"apply", 1, (Blt_Op)ApplyOp, 2, 2, "",},
    {"cancel", 2, (Blt_Op)CancelOp, 2, 2, "",},
    {"cget", 2, (Blt_Op)CgetOp, 3, 3, "value",},
    {"configure", 2, (Blt_Op)ConfigureOp, 2, 0, "?option value...?",},
    {"delete", 1, (Blt_Op)DeleteOp, 3, 4, "first ?last?"},
    {"icursor", 2, (Blt_Op)IcursorOp, 3, 3, "index"},
    {"index", 3, (Blt_Op)IndexOp, 3, 3, "index"},
    {"insert", 3, (Blt_Op)InsertOp, 4, 4, "index string"},
    {"selection", 1, (Blt_Op)SelectionOp, 2, 0, "args"},
};
static int nComboOps = sizeof(comboOps) / sizeof(Blt_OpSpec);

static int
ComboboxCmd(clientData, interp, objc, objv)
    ClientData clientData;
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST *objv;
{
    Combobox *comboPtr = clientData;
    Blt_Op proc;
    int result;

    proc = Blt_GetOpFromObj(interp, nComboOps, comboOps, BLT_OP_ARG1, objc, 
	objv, 0);
    if (proc == NULL) {
	return TCL_ERROR;
    }
    result = (*proc) (comboPtr, interp, objc, objv);
    return result;
}

#endif
